"""This is the ideal user flow
"""
import pytest

from deepeval.metrics.factual_consistency import assert_factual_consistency
from deepeval.metrics.overall_score import OverallScoreMetric
from deepeval.test_case import LLMTestCase
from deepeval.run_test import assert_test, run_test


def generate_llm_output(query: str):
    expected_output = "Our customer success phone line is 1200-231-231."
    return expected_output


def test_llm_output():
    query = "What is the customer success phone line?"
    expected_output = "Our customer success phone line is 1200-231-231."
    output = generate_llm_output(query)
    assert_factual_consistency(output, expected_output)


def test_llm_output_custom():
    query = "What is the customer success phone line?"
    expected_output = "Dogs and cats love to walk around the beach."
    with pytest.raises(AssertionError):
        assert_factual_consistency(query, expected_output)


def test_0():
    query = "How does photosynthesis work?"
    output = "Photosynthesis is the process by which green plants and some other organisms use sunlight to synthesize foods with the help of chlorophyll pigment."
    expected_output = "Photosynthesis is the process by which green plants and some other organisms use sunlight to synthesize food with the help of chlorophyll pigment."
    context = "Biology"

    test_case = LLMTestCase(
        input=query,
        actual_output=output,
        expected_output=expected_output,
        context=context,
    )
    metric = OverallScoreMetric()
    # if you want to make sure that the test returns an error
    assert_test(test_case, metrics=[metric])

    # If you want to run the test
    test_result = run_test(test_case, metrics=[metric])
    # You can also inspect the test result class
    print(test_result[0].success)
    print(test_result[0].score)



import pytest
import shutil
import os
from rich import print
from deepeval.api import Api, TestRun
from typing import Optional, Any
from deepeval.constants import PYTEST_RUN_ENV_VAR, PYTEST_RUN_TEST_NAME
from deepeval.decorators.hyperparameters import get_hyperparameters


def pytest_sessionstart(session: pytest.Session):
    global test_filename
    test_run = TestRun(
        testFile=session.config.getoption("file_or_dir")[0],
        testCases=[],
        metricScores=[],
        configurations={},
    )
    test_filename = test_run.save()


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_protocol(
    item: pytest.Item, nextitem: Optional[pytest.Item]
) -> Optional[Any]:
    os.environ[PYTEST_RUN_TEST_NAME] = item.nodeid.split("::")[-1]
    return None  # continue with the default protocol


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_sessionfinish(session: pytest.Session, exitstatus):
    print("Running teardown with pytest sessionfinish...")
    # Code before yield will run before the test teardown

    # yield control back to pytest for the actual teardown
    yield

    # Code after yield will run after the test teardown
    test_run: TestRun = TestRun.load(test_filename)

    if test_run is None:
        print("Test Run is empty, please try again.")
        return

    if os.getenv(PYTEST_RUN_ENV_VAR) and os.path.exists(".deepeval"):
        api: Api = Api()
        test_run.configurations = get_hyperparameters()
        result = api.post_test_run(test_run)

    # Calculate the average of each metric
    metrics_avg = {
        metric.metric: metric.score for metric in test_run.metric_scores
    }
    # Count the number of passes and failures
    # Get all the possible metrics first
    all_metrics = {metric.metric for metric in test_run.metric_scores}

    # Loop through to filter for each metric
    passes = {
        metric: len(
            [
                test_case_metric
                for test_case in test_run.test_cases
                for test_case_metric in test_case.metrics_metadata
                if test_case_metric.metric == metric and test_case.success
            ]
        )
        for metric in all_metrics
    }
    failures = {
        metric: len(
            [
                test_case_metric
                for test_case in test_run.test_cases
                for test_case_metric in test_case.metrics_metadata
                if test_case_metric.metric == metric
            ]
        )
        - passes[metric]
        for metric in all_metrics
    }
    # Create a table with rich
    from rich.table import Table

    table = Table(title="Test Results")
    table.add_column("Metric", justify="right")
    table.add_column("Average Score", justify="right")
    table.add_column("Passes", justify="right")
    table.add_column("Failures", justify="right")
    table.add_column("Success Rate", justify="right")
    total_passes = 0
    total_failures = 0
    for metric, avg in metrics_avg.items():
        pass_count = passes[metric]
        fail_count = failures[metric]
        total_passes += pass_count
        total_failures += fail_count
        success_rate = pass_count / (pass_count + fail_count) * 100
        table.add_row(
            metric,
            str(avg),
            f"[green]{str(pass_count)}[/green]",
            f"[red]{str(fail_count)}[/red]",
            f"{success_rate:.2f}%",
        )
    total_tests = total_passes + total_failures
    overall_success_rate = total_passes / total_tests * 100
    table.add_row(
        "Total",
        "-",
        f"[green]{str(total_passes)}[/green]",
        f"[red]{str(total_failures)}[/red]",
        f"{overall_success_rate:.2f}%",
    )
    print(table)

    if os.getenv(PYTEST_RUN_ENV_VAR) and os.path.exists(".deepeval"):
        link = f"https://app.confident-ai.com/project/{result.projectId}/unit-tests/{result.testRunId}"
        print(
            "✅ Tests finished! View results on " f"[link={link}]{link}[/link]"
        )
    else:
        print(
            '✅ Tests finished! Run "deepeval login" to view evaluation results in detail.'
        )
    local_folder = os.getenv("DEEPEVAL_RESULTS_FOLDER")
    if local_folder:
        if not os.path.exists(local_folder):
            os.mkdir(local_folder)
            shutil.copy(test_filename, local_folder)
            print(f"Results saved in {local_folder} as {test_filename}")
        elif os.path.isfile(local_folder):
            print(
                f"""❌ Error: DEEPEVAL_RESULTS_FOLDER={local_folder} already exists and is a file.\nDetailed results won't be saved. Please specify a folder or an available path."""
            )
        else:
            shutil.copy(test_filename, local_folder)
            print(f"Results saved in {local_folder} as {test_filename}")
    os.remove(test_filename)


def pytest_terminal_summary(terminalreporter, exitstatus, config):
    for report in terminalreporter.getreports("skipped"):
        if report.skipped:
            reason = report.longreprtext.split("\n")[-1]
            print(f"Test {report.nodeid} was skipped. Reason: {reason}")



"""Overall Score
"""

from deepeval.test_case import LLMTestCase
from deepeval.singleton import Singleton
from deepeval.test_case import LLMTestCase
from deepeval.metrics.answer_relevancy import AnswerRelevancyMetric
from deepeval.metrics.conceptual_similarity import ConceptualSimilarityMetric
from deepeval.metrics.factual_consistency import FactualConsistencyMetric
from deepeval.metrics.base_metric import BaseMetric
from deepeval.run_test import assert_test


class OverallScoreMetric(BaseMetric, metaclass=Singleton):
    def __init__(self, minimum_score: float = 0.5):
        self.minimum_score = minimum_score
        self.answer_relevancy = AnswerRelevancyMetric()
        self.factual_consistency_metric = FactualConsistencyMetric()
        self.conceptual_similarity_metric = ConceptualSimilarityMetric()

    def __call__(self, test_case: LLMTestCase):
        score = self.measure(test_case=test_case)
        self.success = score > self.minimum_score
        return score

    def measure(
        self,
        test_case: LLMTestCase,
    ) -> float:
        metadata = {}
        if test_case.context is not None:
            factual_consistency_score = self.factual_consistency_metric.measure(
                test_case
            )
            metadata["factual_consistency"] = float(factual_consistency_score)

        if test_case.input is not None:
            answer_relevancy_score = self.answer_relevancy.measure(test_case)
            metadata["answer_relevancy"] = float(answer_relevancy_score)

        if test_case.expected_output is not None:
            conceptual_similarity_score = (
                self.conceptual_similarity_metric.measure(test_case)
            )
            metadata["conceptual_similarity"] = float(
                conceptual_similarity_score
            )

        overall_score = sum(metadata.values()) / len(metadata)

        self.success = bool(overall_score > self.minimum_score)
        return overall_score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Overall Score"


def assert_overall_score(
    query: str,
    output: str,
    expected_output: str,
    context: str,
    minimum_score: float = 0.5,
):
    metric = OverallScoreMetric(minimum_score=minimum_score)
    test_case = LLMTestCase(
        input=query,
        actual_output=output,
        expected_output=expected_output,
        context=context,
    )
    assert_test(test_case, metrics=[metric])



from typing import Optional, List
from deepeval.metrics.base_metric import BaseMetric
from deepeval.test_case import LLMTestCase
from deepeval.templates import (
    evaluation_steps_template,
    evaluation_results_template,
)
from deepeval.types import LLMTestCaseParams
from deepeval.chat_completion.retry import call_openai_with_retry
from pydantic import BaseModel
import openai


class LLMEvalMetricResponse(BaseModel):
    score: float


class LLMEvalMetric(BaseMetric):
    def __init__(
        self,
        name: str,
        criteria: str,
        evaluation_params: List[LLMTestCaseParams],
        model: Optional[str] = "gpt-4",
        minimum_score: float = 0.5,
    ):
        self.criteria = criteria
        self.name = name
        self.model = model
        self.evaluation_steps = ""
        self.evaluation_params = evaluation_params

        self.minimum_score = minimum_score

    @property
    def __name__(self):
        return self.name

    def measure(self, test_case: LLMTestCase):
        """LLM evaluated metric based on the GEval framework: https://arxiv.org/pdf/2303.16634.pdf"""

        # Measure the test case
        for param in self.evaluation_params:
            if (
                not hasattr(test_case, param.value)
                or getattr(test_case, param.value) is None
            ):
                raise ValueError(
                    f"Test case is missing the required attribute: {param.value}"
                )

        if self.evaluation_steps == "":
            self.evaluation_steps = self.generate_evaluation_steps()

        score = self.evaluate(test_case)
        score = float(score) * 2 / 10

        self.success = score >= self.minimum_score
        return score

    def is_successful(self):
        return self.success

    def generate_evaluation_steps(self):
        prompt: dict = evaluation_steps_template.format(criteria=self.criteria)

        res = call_openai_with_retry(
            lambda: openai.ChatCompletion.create(
                model=self.model,
                messages=[
                    {
                        "role": "system",
                        "content": "You are a helpful assistant.",
                    },
                    {"role": "user", "content": prompt},
                ],
            )
        )

        return res.choices[0].message.content

    def evaluate(self, test_case: LLMTestCase):
        text = """"""

        for param in self.evaluation_params:
            value = getattr(test_case, param.value)
            text += f"{param.value}: {value} \n\n"

        prompt: dict = evaluation_results_template.format(
            evaluation_steps=self.evaluation_steps,
            text=text,
        )

        res = call_openai_with_retry(
            lambda: openai.ChatCompletion.create(
                model=self.model,
                messages=[{"role": "system", "content": prompt}],
                max_tokens=5,
                top_p=1,
                frequency_penalty=0,
                presence_penalty=0,
                stop=None,
                # logprobs=5,
                n=20,
            )
        )

        total_scores = 0
        count = 0

        for content in res.choices:
            try:
                total_scores += float(content.message.content)
                count += 1
            except:
                pass

        return total_scores / count



import pytest
from deepeval.metrics.factual_consistency import FactualConsistencyMetric
from deepeval.test_case import LLMTestCase
from deepeval.run_test import assert_test


def test_1():
    input = "What does your company do?"
    output = (
        "Specializes in cloud computing, data analytics, and machine learning."
    )
    context = "Our company specializes in cloud computing, data analytics, and machine learning. We offer a range of services including cloud storage solutions, data analytics platforms, and custom machine learning models."
    factual_consistency_metric = FactualConsistencyMetric(minimum_score=1.0)
    test_case = LLMTestCase(input=input, actual_output=output, context=context)
    with pytest.raises(AssertionError):
        assert_test(test_case, [factual_consistency_metric])


def test_2():
    input = "What does your company do?"
    output = (
        "Specializes in cloud computing, data analytics, and machine learning."
    )
    context = "Our company specializes in cloud computing, data analytics, and machine learning. We offer a range of services including cloud storage solutions, data analytics platforms, and custom machine learning models."
    factual_consistency_metric_half = FactualConsistencyMetric(
        minimum_score=0.5
    )
    if factual_consistency_metric_half.minimum_score != 0.5:
        raise ValueError("Minimum score should be 0.5")
    test_case = LLMTestCase(input=input, actual_output=output, context=context)
    assert_test(test_case, [factual_consistency_metric_half])



from abc import abstractmethod

from deepeval.test_case import LLMTestCase


class BaseMetric:
    # set an arbitrary minimum score that will get over-ridden later
    score: float = 0

    @property
    def minimum_score(self) -> float:
        return self._minimum_score

    @minimum_score.setter
    def minimum_score(self, value: float):
        self._minimum_score = value

    # Measure function signature is subject to be different - not sure
    # how applicable this is - might need a better abstraction
    @abstractmethod
    def measure(self, test_case: LLMTestCase, *args, **kwargs):
        raise NotImplementedError

    def _get_init_values(self):
        # We use this method for sending useful metadata
        init_values = {
            param: getattr(self, param)
            for param in vars(self)
            if isinstance(getattr(self, param), (str, int, float))
        }
        return init_values

    @abstractmethod
    def is_successful(self) -> bool:
        raise NotImplementedError

    @property
    def __name__(self):
        return "Base Metric"



"""Module for the query generator
"""
from abc import ABC, abstractmethod
from typing import List


class QueryGenerator(ABC):
    @abstractmethod
    def generate_queries(
        self, texts: List, tags: List = None, num_queries: int = 3
    ):
        pass


class BEIRQueryGenerator:
    """BEIR Query Generator"""

    def __init__(self):
        self._load_model()

    def generate_queries(
        self, texts: List, tags: List = None, num_queries: int = 3
    ):
        """Generates a number of queries"""
        import torch

        queries = []
        for t in texts:
            input_ids = self.tokenizer.encode(t, return_tensors="pt")
            with torch.no_grad():
                outputs = self.model.generate(
                    input_ids=input_ids,
                    max_length=64,
                    do_sample=True,
                    top_p=0.95,
                    num_return_sequences=num_queries,
                )

        for i in range(len(outputs)):
            query = self.tokenizer.decode(outputs[i], skip_special_tokens=True)
            queries.append(query)
        return queries

    def _load_model(self):
        if hasattr(self, "model"):
            return
        try:
            from transformers import T5ForConditionalGeneration, T5Tokenizer
        except ModuleNotFoundError as e:
            raise Exception("Run `pip install transformers`")

        self.tokenizer = T5Tokenizer.from_pretrained(
            "BeIR/query-gen-msmarco-t5-large-v1"
        )
        self.model = T5ForConditionalGeneration.from_pretrained(
            "BeIR/query-gen-msmarco-t5-large-v1"
        )
        self.model.eval()



"""Asserting conceptual similarity
"""
from typing import Optional

from deepeval.singleton import Singleton
from deepeval.test_case import LLMTestCase
from deepeval.utils import cosine_similarity
from deepeval.run_test import assert_test
from deepeval.progress_context import progress_context
from deepeval.metrics.base_metric import BaseMetric


class ConceptualSimilarityMetric(BaseMetric, metaclass=Singleton):
    """basic implementation of ConceptualSimilarityMetric"""

    def __init__(
        self,
        model_name: Optional[str] = "sentence-transformers/all-mpnet-base-v2",
        minimum_score: float = 0.7,
    ):
        from sentence_transformers import SentenceTransformer

        self.model_name = model_name
        with progress_context(
            "Downloading Conceptual Similarity model (may take up to 2 minutes if running for the first time)..."
        ):
            self.model = SentenceTransformer(self.model_name).eval()
        self.minimum_score = minimum_score

    def _vectorize(self, text_a: str, text_b: str):
        vectors = self.model.encode([text_a, text_b])
        return vectors

    def measure(self, test_case: LLMTestCase) -> float:
        if test_case.actual_output is None or test_case.expected_output is None:
            raise ValueError("Output or expected_output cannot be None")

        vectors = self._vectorize(
            test_case.actual_output, test_case.expected_output
        )
        self.score = cosine_similarity(vectors[0], vectors[1])
        return float(self.score)

    def is_successful(self) -> bool:
        return bool(self.score >= self.minimum_score)

    @property
    def __name__(self):
        return "Conceptual Similarity"


def assert_conceptual_similarity(
    output: str, expected_output: str, minimum_score=0.7
):
    metric = ConceptualSimilarityMetric(minimum_score=minimum_score)
    test_case = LLMTestCase(
        input="placeholder",
        actual_output=output,
        expected_output=expected_output,
    )
    assert_test(test_case, [metric])


def is_conceptually_similar(
    output: str, expected_output: str, minimum_score=0.7
) -> bool:
    metric = ConceptualSimilarityMetric(minimum_score=minimum_score)
    test_case = LLMTestCase(
        input="placeholder",
        actual_output=output,
        expected_output=expected_output,
    )
    return metric.measure(test_case) >= minimum_score



import pytest
from deepeval.test_case import LLMTestCase
from deepeval.run_test import assert_test
from deepeval.metrics.factual_consistency import FactualConsistencyMetric
from deepeval.metrics.answer_relevancy import AnswerRelevancyMetric

# This tutorial will guide you on how to use pytest for testing chatbot responses.
# pytest is a testing framework that allows you to easily create small, simple tests, yet scales to support complex functional testing.
# We will use the deepeval library which provides tools for evaluating language models.
# The library includes test cases, metrics, and a function to assert the test results.


# First, we will define a simple test case. The test case includes a query, expected output, and context.
# The query is the question asked to the chatbot, the output is the expected answer, and the context is the actual answer.
# We then use the assert_test function from deepeval to check if the chatbot's response is as expected.
def test_customer_chatbot_simple():
    input = "What are your operating hours?"
    output = "Our operating hours are from 9 AM to 5 PM, Monday to Friday."
    context = "Our company operates from 10 AM to 6 PM, Monday to Friday."
    factual_consistency_metric = FactualConsistencyMetric(minimum_score=0.3)
    answer_relevancy_metric = AnswerRelevancyMetric(minimum_score=0.5)
    test_case = LLMTestCase(input=input, actual_output=output, context=context)
    assert_test(
        test_case, [factual_consistency_metric, answer_relevancy_metric]
    )


# We can also define multiple test cases at once in a list of dictionaries.
# Each dictionary represents a test case and can be directly used in our tests.
# The keys of the dictionary are the column names in the DataFrame (query, output, context).
# The values are the corresponding values for each test case.
CHATBOT_TEST_CASES = [
    {
        "input": "What are your operating hours?",
        "output": "Our operating hours are from 9 AM to 5 PM, Monday to Friday.",
        "context": [
            "Our company operates from 10 AM to 6 PM, Monday to Friday.",
            "We are closed on weekends and public holidays.",
            "Our customer service is available 24/7.",
        ],
    },
    {
        "input": "Do you offer free shipping?",
        "output": "Yes, we offer free shipping on orders over $50.",
        "context": [
            "Our company offers free shipping on orders over $100.",
            "Free shipping applies to all products.",
            "The free shipping offer is valid only within the country.",
        ],
    },
    {
        "input": "What is your return policy?",
        "output": "We accept returns within 30 days of purchase.",
        "context": [
            "Our company accepts returns within 60 days of purchase.",
            "The product must be in its original condition.",
            "The return shipping cost will be covered by the customer.",
        ],
    },
    {
        "input": "Do you have a physical store?",
        "output": "Yes, we have a physical store in New York.",
        "context": [
            "Our company has physical stores in several locations across the country.",
            "The New York store is our flagship store.",
            "Our stores offer a wide range of products.",
        ],
    },
    {
        "input": "Do you offer international shipping?",
        "output": "Yes, we offer international shipping to selected countries.",
        "context": [
            "Our company offers international shipping to all countries.",
            "International shipping rates apply.",
            "Delivery times vary depending on the destination.",
        ],
    },
    {
        "input": "Do you have a customer loyalty program?",
        "output": "Yes, we have a loyalty program called 'Rewards Club'.",
        "context": [
            "Our company has a loyalty program that offers exclusive benefits to members.",
            "The 'Rewards Club' program offers discounts, early access to sales, and more.",
            "Customers can join the 'Rewards Club' by signing up on our website.",
        ],
    },
]


# pytest provides a decorator called 'parametrize' that allows you to run a test function multiple times with different arguments.
# Here, we use it to run the test function for each test case in CHATBOT_TEST_CASES.
# The test function takes a test case as an argument, extracts the query, output, and context, and then runs the test.
@pytest.mark.parametrize(
    "test_case",
    CHATBOT_TEST_CASES,
)
def test_customer_chatbot(test_case: dict):
    input = test_case["input"]
    output = test_case["output"]
    context = test_case["context"]
    factual_consistency_metric = FactualConsistencyMetric(minimum_score=0.3)
    answer_relevancy_metric = AnswerRelevancyMetric(minimum_score=0.5)
    test_case = LLMTestCase(input=input, actual_output=output, context=context)
    assert_test(
        test_case, [factual_consistency_metric, answer_relevancy_metric]
    )


# # You can also load this in a CSV

import pandas as pd

# Here we are defining a CSV formatted string that contains our test cases.
# Each line represents a test case with a query, expected output, and context.
# The context is a string of sentences separated by '|'.
CSV_DATA = """query,output,context
"What are your operating hours?","Our operating hours are from 9 AM to 5 PM, Monday to Friday.","Our company operates from 10 AM to 6 PM, Monday to Friday.|We are closed on weekends and public holidays.|Our customer service is available 24/7."
"What are your operating hours?","We are open from 10 AM to 6 PM, Monday to Friday.","Our company operates from 10 AM to 6 PM, Monday to Friday.|We are closed on weekends and public holidays.|Our customer service is available 24/7."
"Do you offer free shipping?","Yes, we offer free shipping on orders over $50.","Our company offers free shipping on orders over $100.|Free shipping applies to all products.|The free shipping offer is valid only within the country."
"""

# We read the CSV_DATA into a pandas DataFrame.
# This allows us to easily manipulate and process the data.
import io

# Create a temporary file with CSV_DATA
temp_file = io.StringIO(CSV_DATA)

# Read the temporary file as a CSV
df = pd.read_csv(temp_file)

# We then split the context column into a list of sentences.
# This is done by splitting the string on each '|'.
# The result is a list of context sentences for each test case.
df["context"] = df["context"].apply(lambda x: x.split("|"))

# Finally, we convert the DataFrame to a list of dictionaries.
# Each dictionary represents a test case and can be directly used in our tests.
# The keys of the dictionary are the column names in the DataFrame (query, output, context).
# The values are the corresponding values for each test case.
CHATBOT_TEST_CASES = df.to_dict("records")



import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics.llm_eval_metric import LLMEvalMetric
from deepeval.types import LLMTestCaseParams


def test_chat_completion():
    """Test Chat Completion"""
    metric = LLMEvalMetric(
        name="Validity",
        criteria="The response is a valid response to the prompt.",
        minimum_score=0.5,
        evaluation_params=[
            LLMTestCaseParams.INPUT,
            LLMTestCaseParams.ACTUAL_OUTPUT,
        ],
    )
    test_case = LLMTestCase(
        input="What is the capital of France?",
        actual_output="Paris",
        expected_output="Paris",
        context="Geography",
    )
    metric.measure(test_case)
    assert metric.is_successful() is True
    assert metric.measure(test_case) == 1.0



import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics.ragas_metric import RagasMetric
from deepeval.run_test import assert_test

query = "Who won the FIFA World Cup in 2018?"
output = "Winners of the FIFA world cup were the French national football team"
expected_output = "French national football team"
context = "The FIFA World Cup in 2018 was won by the French national football team. They defeated Croatia 4-2 in the final match to claim the championship."


@pytest.mark.skip(reason="openai is expensive")
def test_ragas_score():
    test_case = LLMTestCase(
        input=query,
        actual_output=output,
        expected_output=expected_output,
        context=context,
    )
    metric = RagasMetric()

    with pytest.raises(AssertionError):
        assert_test(
            test_cases=[test_case],
            metrics=[metric],
        )



"""File for handling API key
"""
import json

from .constants import IMPLEMENTATION_ID_NAME, KEY_FILE


class KeyFileHandler:
    def __init__(self):
        self.data = {}

    def write_data(self, key, value):
        """Writes data to the hidden file"""
        self.data[key] = value
        with open(KEY_FILE, "w") as f:
            json.dump(self.data, f)

    def fetch_data(self, key):
        """Fetches the data from the hidden file"""
        try:
            with open(KEY_FILE, "r") as f:
                self.data = json.load(f)
        except FileNotFoundError:
            # Handle the case when the file doesn't exist
            self.data = {}
        return self.data.get(key)

    def write_api_key(self, api_key):
        """Writes the API key to the hidden file"""
        self.write_data("api_key", api_key)

    def fetch_api_key(self) -> str:
        """Fetches the API key from the hidden file"""
        return self.fetch_data("api_key")

    def write_implementation_name(self, implementation_name: str):
        """Writes the implementation name to the hidden file"""
        self.write_data(IMPLEMENTATION_ID_NAME, implementation_name)

    def fetch_implementation_name(self) -> str:
        """Fetches the implementation name from the hidden file"""
        return self.fetch_data(IMPLEMENTATION_ID_NAME)


KEY_FILE_HANDLER = KeyFileHandler()



"""Investigate test case.
"""
import hashlib
from dataclasses import dataclass
from typing import Any, List, Optional, Union


@dataclass
class TestCase:
    id: Optional[str] = None

    def __post_init__(self):
        if self.id is None:
            id_string = "".join(str(value) for value in self.__dict__.values())
            self.id = hashlib.md5(id_string.encode()).hexdigest()


@dataclass
class LLMTestCase(TestCase):
    def __init__(
        self,
        input: str,
        actual_output: str,
        expected_output: Optional[str] = None,
        context: Optional[Union[str, List[str]]] = None,
        retrieval_context: Optional[List[str]] = None,
        id: Optional[str] = None,
    ):
        super().__init__(id)
        self.input = input
        self.actual_output = actual_output
        self.expected_output = expected_output
        self.retrieval_context = retrieval_context
        # Force context to be a list
        if isinstance(context, str):
            context = [context]
        self.context = context

    def __post_init__(self):
        super().__post_init__()
        self.__name__ = f"LLMTestCase_{self.id}"


class AgentTestCase(TestCase):
    """Test Case For Agents"""

    pass



from typing import Callable, Any
import openai
import time


def call_openai_with_retry(
    callable: Callable[[], Any], max_retries: int = 2
) -> Any:
    if not openai.api_key:
        raise ValueError(
            "OpenAI API key is not set. Please ensure it's set in your environment variables or passed explicitly."
        )

    for _ in range(max_retries):
        try:
            response = callable()
            return response
        except openai.error.OpenAIError as e:
            print(f"An error occurred: {e}. Retrying...")
            time.sleep(2)
            continue

    raise Exception(
        "Max retries reached. Unable to make a successful API call to OpenAI."
    )



import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics.factual_consistency import (
    FactualConsistencyMetric,
    assert_factual_consistency,
)
from deepeval.run_test import assert_test


def test_factual_consistency():
    # legacy functions - consider removing
    with pytest.raises(AssertionError):
        assert_factual_consistency(
            context="After a long day at work, Sarah decided to go for a walk in the park to unwind. She put on her sneakers and grabbed her headphones before heading out. As she strolled along the path, she noticed families having picnics, children playing on the playground, and ducks swimming in the pond.",
            output="Sarah spent the evening at the library, engrossed in a book.",
        )


def test_factual_consistency_2():
    # legacy functions - consider removing
    assert_factual_consistency(
        context="After a long day at work, Sarah decided to go for a walk in the park to unwind. She put on her sneakers and grabbed her headphones before heading out. As she strolled along the path, she noticed families having picnics, children playing on the playground, and ducks swimming in the pond.",
        output="Sarah went out for a walk in the park.",
    )


def test_factual_consistency_metric():
    metric = FactualConsistencyMetric(minimum_score=0.8)
    test_case = LLMTestCase(
        input="placeholder",
        actual_output="Python is a programming language.",
        context="Python is a high-level, versatile, and interpreted programming language known for its simplicity and readability.",
    )
    assert_test(test_case, [metric])


def test_factual_consistency_metric_2():
    metric = FactualConsistencyMetric(minimum_score=0.6)
    test_case = LLMTestCase(
        input="placeholder",
        actual_output="Python is a programming language.",
        context="Python is NOT a programming language.",
    )
    with pytest.raises(AssertionError):
        assert_test(test_case, [metric])


def test_factual_consistency_metric_3():
    metric = FactualConsistencyMetric(minimum_score=0.6)
    test_case = LLMTestCase(
        input="placeholder",
        actual_output="Python is a programming language.",
        context="Python is a snake.",
    )
    with pytest.raises(AssertionError):
        assert_test(test_case, [metric])



"""Tests for toxic
"""


import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics.bias_classifier import UnBiasedMetric, assert_unbiased
from deepeval.run_test import assert_test
from deepeval.types import LLMTestCaseParams

output = "Winners of the FIFA world cup were the French national football team"

metric = UnBiasedMetric(evaluation_params=[LLMTestCaseParams.INPUT])


def test_non_toxic():
    assert_test(
        test_cases=LLMTestCase(
            input="placeholder",
            actual_output=output,
        ),
        metrics=[metric],
    )


def test_non_toxic_2():
    with pytest.raises(AssertionError):
        assert_unbiased(
            evaluation_params=[LLMTestCaseParams.ACTUAL_OUTPUT],
            input="placeholder",
            actual_output="Left wing is evil.",
        )


def test_non_toxic_metric():
    metric = UnBiasedMetric(evaluation_params=[LLMTestCaseParams.INPUT])
    test_case = LLMTestCase(input="placeholder", actual_output=output)
    assert_test(test_case, [metric])


def test_bias_score():
    metric = UnBiasedMetric(
        evaluation_params=[
            LLMTestCaseParams.INPUT,
            LLMTestCaseParams.ACTUAL_OUTPUT,
        ]
    )
    test_case = LLMTestCase(input="placeholder", actual_output=output)
    assert_test(test_case, [metric])



from enum import Enum


class LLMTestCaseParams(Enum):
    INPUT = "input"
    ACTUAL_OUTPUT = "actual_output"
    EXPECTED_OUTPUT = "expected_output"
    CONTEXT = "context"
    RETRIEVAL_CONTEXT = "retrieval_context"
    ID = "id"



from deepeval.run_test import TestResult


def test_result():
    result = TestResult(
        success=True,
        score=2.0,
        metric_name="test_metric",
        query="test_query",
        output="test_output",
        expected_output="test_expected_output",
        metadata=None,
        context="test_context",
    )
    assert result.score == 1.0



import os
import platform
import urllib.parse
import requests
import json
import warnings
from collections import defaultdict

from typing import Any, Optional
from pydantic import BaseModel, Field
from typing import List
from requests.adapters import HTTPAdapter, Response, Retry

from deepeval.constants import (
    API_KEY_ENV,
    PYTEST_RUN_ENV_VAR,
    PYTEST_RUN_TEST_NAME,
)
from deepeval.key_handler import KEY_FILE_HANDLER
from deepeval.metrics.base_metric import BaseMetric
from deepeval.test_case import LLMTestCase

API_BASE_URL = "https://app.confident-ai.com/api"
# API_BASE_URL = "http://localhost:3000/api"

# Parameters for HTTP retry
HTTP_TOTAL_RETRIES = 3  # Number of total retries
HTTP_RETRY_BACKOFF_FACTOR = 2  # Wait 1, 2, 4 seconds between retries
HTTP_STATUS_FORCE_LIST = [408, 429] + list(range(500, 531))
HTTP_RETRY_ALLOWED_METHODS = frozenset({"GET", "POST", "DELETE"})


class MetricsMetadata(BaseModel):
    metric: str
    score: float
    minimum_score: float = Field(None, alias="minimumScore")


class APITestCase(BaseModel):
    name: str
    input: str
    actual_output: str = Field(..., alias="actualOutput")
    expected_output: Optional[str] = Field(None, alias="expectedOutput")
    success: bool
    metrics_metadata: List[MetricsMetadata] = Field(
        ..., alias="metricsMetadata"
    )
    threshold: float
    run_duration: float = Field(..., alias="runDuration")
    context: Optional[list] = Field(None)


class MetricScore(BaseModel):
    metric: str
    score: float

    @classmethod
    def from_metric(cls, metric: BaseMetric):
        return cls(metric=metric.__name__, score=metric.score)


class TestRunResponse(BaseModel):
    """Add Test Run Results"""

    testRunId: str
    projectId: str


class MetricDict:
    def __init__(self):
        self.metric_dict = {}
        self.metric_count = {}

    def add_metric(self, metric_name, score):
        if metric_name not in self.metric_dict:
            self.metric_dict[metric_name] = score
            self.metric_count[metric_name] = 1
        else:
            self.metric_dict[metric_name] += score
            self.metric_count[metric_name] += 1

    def get_average_metric_score(self):
        return [
            MetricScore(
                metric=metric,
                score=self.metric_dict[metric] / self.metric_count[metric],
            )
            for metric in self.metric_dict
        ]


class MetricsMetadataAverageDict:
    def __init__(self):
        self.metric_dict = defaultdict(list)
        self.min_score_dict = defaultdict(float)

    def add_metric(self, metric: BaseMetric):
        self.metric_dict[metric.__name__].append(metric.score)
        self.min_score_dict[metric.__name__] = min(
            self.min_score_dict.get(metric.__name__, float("inf")),
            metric.minimum_score,
        )

    def get_metrics_metadata(self):
        return [
            MetricsMetadata(
                metric=metric_name,
                score=sum(scores) / len(scores),
                minimumScore=self.min_score_dict[metric_name],
            )
            for metric_name, scores in self.metric_dict.items()
        ]


class TestRun(BaseModel):
    test_file: Optional[str] = Field(
        # TODO: Fix test_file
        "test.py",
        alias="testFile",
    )
    test_cases: List[APITestCase] = Field(
        alias="testCases", default_factory=lambda: []
    )
    metric_scores: List[MetricScore] = Field(
        default_factory=lambda: [], alias="metricScores"
    )
    configurations: dict

    def add_llm_test_case(
        self,
        test_case: LLMTestCase,
        metrics: List[BaseMetric],
        run_duration: float,
    ):
        # Check if test case with the same ID already exists
        # TODO: bug for pytest batch runs - unable to find test case name
        existing_test_case: APITestCase = next(
            (tc for tc in self.test_cases if tc.name == test_case.__name__),
            None,
        )

        metrics_metadata_dict = MetricsMetadataAverageDict()
        for metric in metrics:
            metrics_metadata_dict.add_metric(metric)
        metrics_metadata = metrics_metadata_dict.get_metrics_metadata()
        success = all([metric.is_successful() for metric in metrics])
        threshold = metrics[0].minimum_score

        if existing_test_case:
            # If it exists, append the metrics to the existing test case
            existing_test_case.metrics_metadata.extend(metrics_metadata)
            # Update the success status and threshold
            existing_test_case.success = success
            existing_test_case.threshold = threshold
        else:
            # If it doesn't exist, create a new test case
            # Adding backwards compatibility to ensure context still works.
            context = test_case.context
            if isinstance(context, str):
                context = [context]
            self.test_cases.append(
                APITestCase(
                    # Get the test from the pytest plugin
                    name=os.getenv(PYTEST_RUN_TEST_NAME, "-"),
                    input=test_case.input,
                    actualOutput=test_case.actual_output,
                    expectedOutput=test_case.expected_output,
                    success=success,
                    metricsMetadata=metrics_metadata,
                    threshold=threshold,
                    runDuration=run_duration,
                    context=context,
                )
            )

        all_metric_dict = MetricDict()

        for test_case in self.test_cases:
            test_case: APITestCase
            metrics = test_case.metrics_metadata
            for metric in metrics:
                metric: MetricsMetadata
                all_metric_dict.add_metric(metric.metric, metric.score)

        self.metric_scores = all_metric_dict.get_average_metric_score()

    def save(self, file_path: Optional[str] = None):
        if file_path is None:
            file_path = os.getenv(PYTEST_RUN_ENV_VAR)
            # If file Path is None, remove it
            if not file_path:
                return
            elif not file_path.endswith(".json"):
                file_path = f"{file_path}.json"
        with open(file_path, "w") as f:
            json.dump(self.dict(by_alias=True, exclude_none=True), f)

        return file_path

    @classmethod
    def load(cls, file_path: Optional[str] = None):
        if file_path is None:
            file_path = os.getenv(PYTEST_RUN_ENV_VAR)
            # If file Path is None, remove it
            if not file_path:
                return
            elif not file_path.endswith(".json"):
                file_path = f"{file_path}.json"
        with open(file_path, "r") as f:
            return cls(**json.load(f))


class Api:
    """Internal Api reference for handling http operations"""

    def __init__(
        self,
        api_key: str = os.getenv(API_KEY_ENV, ""),
        user_agent_extension=None,
        api_instance_url=None,
        verify=None,
        proxies=None,
        cert=None,
    ):
        if api_key == "":
            # get API key if none is supplied after you log in
            api_key = KEY_FILE_HANDLER.fetch_api_key()

        if api_key == "" or api_key is None:
            raise ValueError("Please provide a valid API Key.")

        self.api_key = api_key

        self._auth = (self.api_key, "")
        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": self._generate_useragent(user_agent_extension),
            # "Authorization": "Bearer " + api_key,
            # This is what gets sent now - auth gets sent to firebase instead
            "CONFIDENT_API_KEY": api_key,
        }
        self._headers_multipart_form_data = {
            "User-Agent": self._generate_useragent(user_agent_extension),
        }
        self.base_api_url = api_instance_url or API_BASE_URL

        self.verify = verify
        self.proxies = proxies
        self.cert = cert

    @staticmethod
    def _http_request(
        method,
        url,
        headers=None,
        auth=None,
        params=None,
        body=None,
        files=None,
        data=None,
        verify=None,
        proxies=None,
        cert=None,
    ) -> Response:
        https = requests.Session()
        retry_strategy = Retry(
            total=HTTP_TOTAL_RETRIES,
            backoff_factor=HTTP_RETRY_BACKOFF_FACTOR,
            status_forcelist=HTTP_STATUS_FORCE_LIST,
            allowed_methods=HTTP_RETRY_ALLOWED_METHODS,
            raise_on_status=False,
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)
        https.mount("https://", adapter)

        https.cert = cert if cert else None
        https.verify = verify if verify else True
        if proxies:
            https.proxies.update(proxies)

        try:
            params = params or {}
            body = body or None

            res = https.request(
                method=method,
                url=url,
                headers=headers,
                # auth=auth,
                params=params,
                json=body,
                files=files,
                data=data,
            )

            return res
        except Exception as err:
            raise Exception(err) from err

    @staticmethod
    def _raise_on_response(res: Response):
        try:
            message = res.json().get("error", res.text)
        except ValueError:
            message = res.text
        if res.status_code == 410:
            warnings.warn(f"Deprecation Warning: {message}", DeprecationWarning)
        raise Exception(message)

    def _api_request(
        self,
        method,
        endpoint,
        headers=None,
        auth=None,
        params=None,
        body=None,
        files=None,
        data=None,
    ):
        """Generic HTTP request method with error handling."""

        url = f"{self.base_api_url}/{endpoint}"
        res = self._http_request(
            method,
            url,
            headers,
            auth,
            params,
            body,
            files,
            data,
            verify=self.verify,
            proxies=self.proxies,
            cert=self.cert,
        )

        json = None
        if res.status_code == 200:
            try:
                json = res.json()
            except ValueError:
                # Some endpoints only return 'OK' message without JSON
                return json
        elif (
            res.status_code == 409
            and "task" in endpoint
            and body.get("unique_id")
        ):
            retry_history = res.raw.retries.history
            # Example RequestHistory tuple
            # RequestHistory(method='POST',
            #   url='/v1/task/imageannotation',
            #   error=None,
            #   status=409,
            #   redirect_location=None)
            if retry_history != ():
                # See if the first retry was a 500 or 503 error
                if retry_history[0][3] >= 500:
                    uuid = body["unique_id"]
                    newUrl = f"{self.base_api_url}/tasks?unique_id={uuid}"
                    # grab task from api
                    newRes = self._http_request(
                        "GET", newUrl, headers=headers, auth=auth
                    )
                    json = newRes.json()["docs"][0]
            else:
                self._raise_on_response(res)
        else:
            self._raise_on_response(res)
        return json

    def get_request(self, endpoint, params=None):
        """Generic GET Request Wrapper"""
        return self._api_request(
            "GET",
            endpoint,
            headers=self._headers,
            auth=self._auth,
            params=params,
        )

    def post_request(self, endpoint, body=None, files=None, data=None):
        """Generic POST Request Wrapper"""
        return self._api_request(
            "POST",
            endpoint,
            headers=self._headers
            if files is None
            else self._headers_multipart_form_data,
            auth=self._auth,
            body=body,
            files=files,
            data=data,
        )

    def delete_request(self, endpoint, params=None, body=None):
        """Generic DELETE Request Wrapper"""
        return self._api_request(
            "DELETE",
            endpoint,
            headers=self._headers,
            auth=self._auth,
            params=params,
            body=body,
        )

    def put_request(self, endpoint, body=None, params=None):
        """Generic PUT Request Wrapper"""
        return self._api_request(
            "PUT",
            endpoint,
            body=body,
            headers=self._headers,
            auth=self._auth,
            params=params,
        )

    @staticmethod
    def _generate_useragent(extension: str = None) -> str:
        """Generates UserAgent parameter with module, Python
        and OS details

        Args:
            extension (str, optional): Option to extend UserAgent
            with source system

        Returns:
            str: Generated UserAgent parameter with platform versions
        """
        python_version = platform.python_version()
        os_platform = platform.platform()

        user_agent = " ".join(
            filter(
                None,
                [
                    f"Python/{python_version}",
                    f"OS/{os_platform}",
                    extension,
                ],
            )
        )
        return user_agent

    @staticmethod
    def quote_string(text: str) -> str:
        """Project and Batch names can be a part of URL, which causes
        an error in case of a special character used.
        Quotation assures the right object to be retrieved from API.

        `quote_string('a bc/def')` -> `a%20bc%2Fdef`

        Args:
            text (str): Input text to be quoted

        Returns:
            str: Quoted text in return
        """
        return urllib.parse.quote(text, safe="")

    def add_test_case(
        self,
        query: str,
        actual_output: str,
        metric_score: float,
        metric_name: str,
        success: bool,
        datapoint_id: str,
        implementation_id: str,
        metrics_metadata: Any,
    ):
        """send test case data to the prod-data endpoint"""
        if not metrics_metadata:
            metrics_metadata = {}
        return self.post_request(
            endpoint="/v1/prod-data",
            body={
                "query": query,
                "actualOutput": actual_output,
                "defaultMetricScore": metric_score,
                "defaultMetricName": metric_name,
                "implementationId": implementation_id,
                "success": success,
                "metricsMetadata": metrics_metadata,
                "goldenId": datapoint_id,
            },
        )

    def add_golden(
        self,
        query: str = "-",
        expected_output: str = "-",
        context: str = "-",
        is_synthetic: bool = False,
    ):
        return self.post_request(
            endpoint="v1/golden",
            body={
                "query": query,
                "expectedOutput": expected_output,
                "context": context,
                "isSynthetic": is_synthetic,
            },
        )

    def list_prod_data(self, implementation_id: str):
        return self.get_request(
            endpoint="/v1/prod-data?implementationId=" + implementation_id
        )

    def create_implementation(
        self, name: str, description: Optional[str] = None
    ):
        body = {"name": name}
        if description:
            body["description"] = description
        return self.post_request(endpoint="/v1/implementation", body=body)

    def list_implementations(self):
        """
        Returns a list of implementations
        """
        return self.get_request(endpoint="/v1/implementation")

    def post_test_run(self, test_run: TestRun) -> TestRunResponse:
        """Post a test run"""
        try:
            # make sure to exclude none for `context` to ensure it is handled properly
            body = test_run.model_dump(by_alias=True, exclude_none=True)
        except AttributeError:
            # Pydantic version below 2.0
            body = test_run.dict(by_alias=True, exclude_none=True)

        result = self.post_request(
            endpoint="/v1/test-run",
            body=body,
        )
        response = TestRunResponse(
            testRunId=result["testRunId"], projectId=result["projectId"]
        )
        return response



import typer
from typing_extensions import Annotated

# Rich has a few dependency issues
try:
    from rich import print
except Exception as e:
    pass
from ..api import Api
from ..constants import IMPLEMENTATION_ID_NAME
from ..key_handler import KEY_FILE_HANDLER
from .test import app as test_app

app = typer.Typer(name="deepeval")

app.add_typer(test_app, name="test")


@app.command()
def login(
    api_key: Annotated[
        str,
        typer.Option(
            help="API key to get from https://app.confident-ai.com. Required if you want to log events to the server."
        ),
    ] = ""
):
    """Login to the DeepEval platform."""
    print("Welcome to [bold]DeepEval[/bold]!")
    print(
        "Grab your API key here: [link=https://app.confident-ai.com]https://app.confident-ai.com[/link] "
    )
    if api_key == "":
        while True:
            api_key = input("Paste your API Key: ").strip()
            if api_key:
                break
            else:
                print("API Key cannot be empty. Please try again.\n")
    KEY_FILE_HANDLER.write_api_key(api_key)
    client = Api(api_key=api_key)
    print("Success! :raising_hands:")
    print(
        "If you are new to DeepEval, try generate a sample test: [bold]deepeval test generate --output-file test_sample.py[/bold]"
    )
    print("Run a sample test: [bold]deepeval test run test_sample.py[/bold]")


@app.command()
def switch(
    implementation_name: Annotated[
        str, typer.Option(help="The name of the project you want to switch to")
    ] = "",
):
    """Switch to a different project on the DeepEval platform."""
    if implementation_name == "":
        print("You must provide a project name to switch to.")
    else:
        KEY_FILE_HANDLER.write_data(IMPLEMENTATION_ID_NAME, implementation_name)
        print(f"Switched to project: {implementation_name}")


if __name__ == "__main__":
    app()



from rouge_score import rouge_scorer
from nltk.tokenize import word_tokenize
from nltk.translate.bleu_score import sentence_bleu
from typing import Union, List, Optional


class StatisticalCalculator:
    """This class calculates various statistical Natural Language Processing (NLP) evaluation metrics."""

    # Todo: More metrics are to be added

    @classmethod
    def rouge_score(
        cls, target: str, prediction: str, score_type: str
    ) -> float:
        """Calculates the Rouge score for a given target and prediction.

        Rouge (Recall-Oriented Understudy for Gisting Evaluation) is a metric used for evaluating the quality of generated text,
        especially in tasks like text summarization.

        Args:
            target (str): The actual label or target text.
            prediction (str): The generated text from the model or LLM.
            score_type (str): The Rouge score type (Options: 'rouge1', 'rouge2', 'rougeL').

        Returns:
            float: The Rouge score for the given target and prediction, based on the specified score type.
        """
        assert score_type in [
            "rouge1",
            "rouge2",
            "rougeL",
        ], "score_type can be either rouge1, rouge2 or rougeL"
        scorer = rouge_scorer.RougeScorer([score_type], use_stemmer=True)
        scores = scorer.score(target, prediction)
        return scores[score_type].fmeasure

    @classmethod
    def sentence_bleu_score(
        cls,
        references: Union[str, List[str]],
        prediction: str,
        bleu_type: Optional[str] = "bleu1",
    ) -> float:
        """Calculates the BLEU (Bilingual Evaluation Understudy) score for a given prediction compared to one or more reference sentences.

        BLEU is a metric used to evaluate the quality of machine-generated text by comparing it to one or more reference sentences.
        It measures the similarity of the generated text to the reference text based on n-grams.

        Args:
            references (Union[str, List[str]): A reference sentence or a list of reference sentences.
            prediction (str): The generated text or sentence to be evaluated.
            bleu_type (Optional[str]): The BLEU score type (Options: 'bleu1', 'bleu2', 'bleu3', 'bleu4'). Default is 'bleu1'.

        Returns:
            float: The BLEU score for the given prediction and references.
        """
        assert bleu_type in [
            "bleu1",
            "bleu2",
            "bleu3",
            "bleu4",
        ], "Invalud bleu_type. Options: 'bleu1', 'bleu2', 'bleu3', 'bleu4'"
        targets = [references] if isinstance(references, str) else references
        tokenized_targets = [word_tokenize(target) for target in targets]
        tokenized_prediction = word_tokenize(prediction)
        bleu_weight_map = {
            "bleu1": (1, 0, 0, 0),
            "bleu2": (0, 1, 0, 0),
            "bleu3": (0, 0, 1, 0),
            "bleu4": (0, 0, 0, 1),
        }
        return sentence_bleu(
            tokenized_targets,
            tokenized_prediction,
            weights=bleu_weight_map[bleu_type],
        )

    @classmethod
    def exact_match(cls, target: str, prediction: str) -> float:
        raise NotImplementedError()

    @classmethod
    def quasi_exact_match(cls, target: str, prediction: str) -> float:
        raise NotImplementedError()


class ModelBasedCalculator:
    """ModelBasedCalculator uses seperate external models to evaluate the LLM's predictions.
    Example: BertScore. This metrics uses BERT model in order to calculate the score.
    """

    # Todo: More metrics to be added

    @classmethod
    def bert_score(cls, target: str, prediction: str) -> float:
        raise NotImplementedError()

    @classmethod
    def faithfulness_score(cls, target: str, prediction: str) -> float:
        raise NotImplementedError()

    @classmethod
    def PII_score(cls, target: str, prediction: str) -> float:
        raise NotImplementedError()

    @classmethod
    def toxic_score(cls, target: str, prediction: str) -> float:
        raise NotImplementedError()



import re
import string
import numpy as np


def softmax(x):
    e_x = np.exp(x - np.max(x, axis=1, keepdims=True))
    return e_x / e_x.sum(axis=1, keepdims=True)


def cosine_similarity(vector_a, vector_b):
    dot_product = np.dot(vector_a, vector_b)
    norm_a = np.linalg.norm(vector_a)
    norm_b = np.linalg.norm(vector_b)
    similarity = dot_product / (norm_a * norm_b)
    return similarity


def chunk_text(text, chunk_size=20):
    words = text.split()
    chunks = [
        " ".join(words[i : i + chunk_size])
        for i in range(0, len(words), chunk_size)
    ]
    return chunks


def normalize_text(text: str) -> str:
    """Lower text and remove punctuation, articles and extra whitespace.
    Copied from the [QuAC](http://quac.ai/) evaluation script found at
    https://s3.amazonaws.com/my89public/quac/scorer.py"""

    def remove_articles(text: str) -> str:
        return re.sub(r"\b(a|an|the)\b", " ", text)

    def white_space_fix(text: str) -> str:
        return " ".join(text.split())

    def remove_punc(text: str) -> str:
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text: str) -> str:
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(text))))
